from cardy import distance
from utils import test


@test
def empty_card_sorts_have_a_distance_of_zero():
    assert distance((), ()) == 0


@test
def equivalent_card_sorts_have_an_edit_distance_of_zero():
    assert distance(({1},), ({1},)) == 0
    assert distance(({1}, {2}), ({1}, {2})) == 0
    assert distance(({2}, {1}), ({1}, {2})) == 0
    assert distance(
        ({1, 2, 3}, {4}, {5, 6}),
        ({4}, {1, 2, 3}, {5, 6}),
    ) == 0


@test
def empty_groups_are_ignored_when_computing_distances():
    assert distance(({1}, {2}, set()), ({1}, {2})) == 0
    assert distance(({1}, {2}), ({1}, set(), {2})) == 0


@test
def single_card_displacements_have_an_edit_distance_of_one():
    assert distance(({1, 2}, {3}), ({1}, {2, 3})) == 1
    assert distance(({1}, {2}, {3}), ({1}, {2, 3})) == 1
    assert distance(({1}, {2, 3}), ({1, 2, 3},)) == 1


@test
def distance_between_card_sorts_is_computed_for_multiple_moves():
    assert distance(
        ({1, 2, 3}, {4, 5, 6}, {7, 8, 9}),
        ({1, 2}, {3, 4}, {5, 6, 7}, {8, 9}),
    ) == 3
    assert distance(
        ({1, 2}, {3, 4}, {5, 6, 7}, {8, 9}),
        ({1, 2, 3}, {4, 5, 6}, {7, 8, 9}),
    ) == 3



from cardy import neighbourhood
from utils import test

SORTS = {
    0: ({1, 2, 3}, {4, 5}),
    1: ({1, 2, 3}, {4, 5}, set()),
    2: ({1, 2}, {3}, {4, 5}),
    3: ({1, 2}, {3, 4}, {5}),
    4: ({1, 2, 4}, {3, 5},),
}


@test
def zero_neighbourhoods_only_include_equivalent_card_sorts():
    assert neighbourhood(0, SORTS[0], SORTS) == {0, 1}
    assert neighbourhood(0, SORTS[2], SORTS) == {2}
    assert neighbourhood(0, ({1, 2, 3, 4, 5},), SORTS) == set()


@test
def neighbourhoods_are_returned_when_sorts_have_different_distances():
    assert neighbourhood(1, SORTS[0], SORTS) == {0, 1, 2}
    assert neighbourhood(2, SORTS[0], SORTS) == {0, 1, 2, 3, 4}
    assert neighbourhood(2, ({1, 2, 3, 4, 5},), SORTS) == {0, 1, 4}
    assert neighbourhood(3, ({1, 2, 3, 4, 5},), SORTS) == {0, 1, 2, 3, 4}



from munkres import Munkres, make_cost_matrix

from .types import CardSort

__all__ = ("distance",)


def distance[T](sort1: CardSort[T], sort2: CardSort[T]) -> int:
    """Computes the edit distance between the two given card sorts."""
    if not sort1 and not sort2:
        return 0

    weights = [
        [len(group1 & group2) for group2 in sort2]
        for group1 in sort1
    ]
    cost_matrix = make_cost_matrix(weights)
    total = sum([
        weights[row][col]
        for row, col in Munkres().compute(cost_matrix)
    ])
    return sum(len(g) for g in sort1) - total



from collections.abc import Callable, Collection, Mapping, Set

__all__ = ("CardSort", "CliqueHeuristic")

type CardSort[T] = Collection[Set[T]]
type CliqueHeuristic[K, T] = \
    Callable[[int, Mapping[K, CardSort[T]]], K]



from collections.abc import Mapping

from .distance import distance
from .types import CardSort

__all__ = ("neighbourhood",)


def neighbourhood[K, T](
      d: int,
      probe: CardSort[T],
      sorts: Mapping[K, CardSort[T]],
) -> set[K]:
    """
    Returns the d-neighbourhood of the given probe sort in the sorts iterable.

    The probe sort does not need to be one of the given sorts and will not be
    included in the result if it is not.

    :param d: The max distance neighbourhood elements and the probe
    :param probe: The sort at the centre of the neighbourhood
    :param sorts: A collection of sorts to search for the neighbourhood in
    :return: The d-neighbourhood of the given probe
    """
    return {key for key, sort in sorts.items() if distance(probe, sort) <= d}
